@@ -1,11 +1,10 @@
 /*
  * Pextlib.c
- * $Id: Pextlib.c 135852 2015-05-05 17:36:35Z cal@macports.org $
  *
  * Copyright (c) 2002 - 2003 Apple Inc.
  * Copyright (c) 2004 - 2005 Paul Guyot <pguyot@kallisys.net>
  * Copyright (c) 2004 Landon Fuller <landonf@macports.org>
- * Copyright (c) 2012 The MacPorts Project
+ * Copyright (c) 2007 - 2016 The MacPorts Project
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -43,7 +42,10 @@
 #define _BSD_SOURCE
 /* required for clearenv(3)/setenv(3)/unsetenv(3) on OS X */
 #define _DARWIN_C_SOURCE
+/* required for vasprintf(3) on Linux */
+#define _GNU_SOURCE
 
+#include <sys/resource.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
@@ -89,7 +91,7 @@
 #include "strsed.h"
 #include "readdir.h"
 #include "pipe.h"
-#include "flock.h"
+#include "adv-flock.h"
 #include "system.h"
 #include "mktemp.h"
 #include "realpath.h"
@@ -105,72 +107,80 @@ extern char **environ;
 #include "setmode.h"
 #endif
 
-static char *
-ui_escape(const char *source)
-{
-    char *d, *dest;
-    const char *s;
-    size_t dlen;
+__attribute__((format(printf, 3, 0)))
+static void ui_message(Tcl_Interp *interp, const char *severity, const char *format, va_list va) {
+    char *tclcmd;
+    char *buf;
 
-    s = source;
-    dlen = strlen(source) * 2 + 1;
-    d = dest = malloc(dlen);
-    if (dest == NULL) {
-        return NULL;
+    if (vasprintf(&buf, format, va) < 0) {
+        perror("vasprintf");
+        return;
     }
-    while(*s != '\0') {
-        switch(*s) {
-            case '\\':
-            case '}':
-            case '{':
-                *d = '\\';
-                d++;
-                *d = *s;
-                d++;
-                s++;
-                break;
-            case '\n':
-                s++;
-                break;
-            default:
-                *d = *s;
-                d++;
-                s++;
-                break;
-        }
+    if (asprintf(&tclcmd, "ui_%s $warn", severity) < 0) {
+        perror("asprintf");
+        free(buf);
+        return;
     }
-    *d = '\0';
-    return dest;
+
+    Tcl_SetVar(interp, "warn", buf, 0);
+    if (TCL_OK != Tcl_EvalEx(interp, tclcmd, -1, 0)) {
+        fprintf(stderr, "Error evaluating Tcl statement '%s': %s (message: '%s')\n", tclcmd, Tcl_GetStringResult(interp), buf);
+    }
+    Tcl_UnsetVar(interp, "warn", 0);
+    free(buf);
+    free(tclcmd);
 }
 
-int
-ui_info(Tcl_Interp *interp, char *mesg)
-{
-    const char ui_proc_start[] = "ui_info [subst -nocommands -novariables {";
-    const char ui_proc_end[] = "}]";
-    char *script, *string;
-    size_t scriptlen, len, remaining;
-    int rval;
+__attribute__((format(printf, 2, 3)))
+void ui_error(Tcl_Interp *interp, const char *format, ...) {
+    va_list va;
+    va_start(va, format);
+    ui_message(interp, "error", format, va);
+    va_end(va);
+}
 
-    string = ui_escape(mesg);
-    if (string == NULL)
-        return TCL_ERROR;
+__attribute__((format(printf, 2, 3)))
+void ui_warn(Tcl_Interp *interp, const char *format, ...) {
+    va_list va;
 
-    len = strlen(string);
-    scriptlen = sizeof(ui_proc_start) + len + sizeof(ui_proc_end) - 1;
-    script = malloc(scriptlen);
-    if (script == NULL)
-        return TCL_ERROR;
+    va_start(va, format);
+    ui_message(interp, "warn", format, va);
+    va_end(va);
+}
 
-    memcpy(script, ui_proc_start, sizeof(ui_proc_start));
-    remaining = scriptlen - sizeof(ui_proc_start);
-    strncat(script, string, remaining);
-    remaining -= len;
-    strncat(script, ui_proc_end, remaining);
-    free(string);
-    rval = Tcl_EvalEx(interp, script, -1, 0);
-    free(script);
-    return rval;
+__attribute__((format(printf, 2, 3)))
+void ui_msg(Tcl_Interp *interp, const char *format, ...) {
+    va_list va;
+    va_start(va, format);
+    ui_message(interp, "msg", format, va);
+    va_end(va);
+}
+
+__attribute__((format(printf, 2, 3)))
+void ui_notice(Tcl_Interp *interp, const char *format, ...) {
+    va_list va;
+
+    va_start(va, format);
+    ui_message(interp, "notice", format, va);
+    va_end(va);
+}
+
+__attribute__((format(printf, 2, 3)))
+void ui_info(Tcl_Interp *interp, const char *format, ...) {
+    va_list va;
+
+    va_start(va, format);
+    ui_message(interp, "info", format, va);
+    va_end(va);
+}
+
+__attribute__((format(printf, 2, 3)))
+void ui_debug(Tcl_Interp *interp, const char *format, ...) {
+    va_list va;
+
+    va_start(va, format);
+    ui_message(interp, "debug", format, va);
+    va_end(va);
 }
 
 int StrsedCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
@@ -201,24 +211,25 @@ int ExistsuserCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, Tc
 {
     Tcl_Obj *tcl_result;
     struct passwd *pwent;
-    char *user;
+    const char *user;
 
     if (objc != 2) {
         Tcl_WrongNumArgs(interp, 1, objv, "user");
         return TCL_ERROR;
     }
 
-    user = strdup(Tcl_GetString(objv[1]));
-    if (isdigit(*(user)))
-        pwent = getpwuid((uid_t)strtol(user, 0, 0));
-    else
+    user = Tcl_GetString(objv[1]);
+    if (isdigit(*user)) {
+        pwent = getpwuid((uid_t) strtol(user, 0, 0));
+    } else {
         pwent = getpwnam(user);
-    free(user);
+    }
 
-    if (pwent == NULL)
+    if (pwent == NULL) {
         tcl_result = Tcl_NewIntObj(-1);
-    else
+    } else {
         tcl_result = Tcl_NewIntObj(pwent->pw_uid);
+    }
 
     Tcl_SetObjResult(interp, tcl_result);
     return TCL_OK;
@@ -228,24 +239,25 @@ int ExistsgroupCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc, T
 {
     Tcl_Obj *tcl_result;
     struct group *grent;
-    char *group;
+    const char *group;
 
     if (objc != 2) {
         Tcl_WrongNumArgs(interp, 1, objv, "groupname");
         return TCL_ERROR;
     }
 
-    group = strdup(Tcl_GetString(objv[1]));
-    if (isdigit(*(group)))
-        grent = getgrgid((gid_t)strtol(group, 0, 0));
-    else
+    group = Tcl_GetString(objv[1]);
+    if (isdigit(*group)) {
+        grent = getgrgid((gid_t) strtol(group, 0, 0));
+    } else {
         grent = getgrnam(group);
-    free(group);
+    }
 
-    if (grent == NULL)
+    if (grent == NULL) {
         tcl_result = Tcl_NewIntObj(-1);
-    else
+    } else {
         tcl_result = Tcl_NewIntObj(grent->gr_gid);
+    }
 
     Tcl_SetObjResult(interp, tcl_result);
     return TCL_OK;
@@ -600,13 +612,47 @@ int CheckBrokenDNSCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc
     return TCL_OK;
 }
 
+/*
+	set_max_open_files
+
+	synopsis: set_max_open_files
+
+	raises the limit of open files to the maximum
+*/
+int SetMaxOpenFilesCmd(ClientData clientData UNUSED, Tcl_Interp *interp, int objc UNUSED, Tcl_Obj *CONST objv[] UNUSED)
+{
+	struct rlimit rl;
+
+    if (getrlimit(RLIMIT_NOFILE, &rl) == -1) {
+        Tcl_SetErrno(errno);
+        Tcl_ResetResult(interp);
+        Tcl_AppendResult(interp, "set_max_open_files: getrlimit failed: ", (char *)Tcl_PosixError(interp), NULL);
+        return TCL_ERROR;
+    } else {
+#ifdef OPEN_MAX
+        if (rl.rlim_max > OPEN_MAX) {
+            rl.rlim_max = OPEN_MAX;
+        }
+#endif
+        rl.rlim_cur = rl.rlim_max;
+        if (setrlimit(RLIMIT_NOFILE, &rl) == -1) {
+            Tcl_SetErrno(errno);
+            Tcl_ResetResult(interp);
+            Tcl_AppendResult(interp, "set_max_open_files: setrlimit failed: ", (char *)Tcl_PosixError(interp), NULL);
+            return TCL_ERROR;
+        }
+    }
+
+	return TCL_OK;
+}
+
 int Pextlib_Init(Tcl_Interp *interp)
 {
     if (Tcl_InitStubs(interp, "8.4", 0) == NULL)
         return TCL_ERROR;
 
 	Tcl_CreateObjCommand(interp, "system", SystemCmd, NULL, NULL);
-	Tcl_CreateObjCommand(interp, "flock", FlockCmd, NULL, NULL);
+	Tcl_CreateObjCommand(interp, "adv-flock", AdvFlockCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "readdir", ReaddirCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "strsed", StrsedCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "mkstemp", MkstempCmd, NULL, NULL);
@@ -620,10 +666,6 @@ int Pextlib_Init(Tcl_Interp *interp)
 	Tcl_CreateObjCommand(interp, "xinstall", InstallCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "fs-traverse", FsTraverseCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "filemap", FilemapCmd, NULL, NULL);
-#if 1
-	/* the name "rpm-vercomp" is deprecated, use "vercmp" instead */
-	Tcl_CreateObjCommand(interp, "rpm-vercomp", VercompCmd, NULL, NULL);
-#endif
 	Tcl_CreateObjCommand(interp, "vercmp", VercompCmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "rmd160", RMD160Cmd, NULL, NULL);
 	Tcl_CreateObjCommand(interp, "sha256", SHA256Cmd, NULL, NULL);
@@ -652,6 +694,7 @@ int Pextlib_Init(Tcl_Interp *interp)
     Tcl_CreateObjCommand(interp, "seteuid", seteuidCmd, NULL, NULL);
     Tcl_CreateObjCommand(interp, "setgid", setgidCmd, NULL, NULL);
     Tcl_CreateObjCommand(interp, "setegid", setegidCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "getpwuid", getpwuidCmd, NULL, NULL);
     Tcl_CreateObjCommand(interp, "name_to_uid", name_to_uidCmd, NULL, NULL);
     Tcl_CreateObjCommand(interp, "uid_to_name", uid_to_nameCmd, NULL, NULL);
     Tcl_CreateObjCommand(interp, "uname_to_gid", uname_to_gidCmd, NULL, NULL);
@@ -663,6 +706,7 @@ int Pextlib_Init(Tcl_Interp *interp)
     Tcl_CreateObjCommand(interp, "term_get_size", TermGetSizeCmd, NULL, NULL);
 
     Tcl_CreateObjCommand(interp, "check_broken_dns", CheckBrokenDNSCmd, NULL, NULL);
+    Tcl_CreateObjCommand(interp, "set_max_open_files", SetMaxOpenFilesCmd, NULL, NULL);
 
     if (Tcl_PkgProvide(interp, "Pextlib", "1.0") != TCL_OK)
         return TCL_ERROR;
