@@ -2,10 +2,9 @@
  */
 /*
  * tracelib.c
- * $Id$
  *
  * Copyright (c) 2007-2008 Eugene Pimenov (GSoC)
- * Copyright (c) 2008-2010, 2012-2013 The MacPorts Project
+ * Copyright (c) 2008-2010, 2012-2015 The MacPorts Project
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -39,9 +38,11 @@
 
 #include <errno.h>
 #include <fcntl.h>
-#include <limits.h>
+#include <inttypes.h>
 #include <pthread.h>
+#include <signal.h>
 #include <stdarg.h>
+#include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -49,7 +50,6 @@
 #if HAVE_SYS_EVENT_H
 #include <sys/event.h>
 #endif
-#include <sys/resource.h>
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -61,8 +61,15 @@
 #include <registry2.0/registry.h>
 #include <darwintracelib1.0/sandbox_actions.h>
 
+#if defined(LOCAL_PEERPID) && defined(HAVE_LIBPROC_H)
+#include <libproc.h>
+#define HAVE_PEERPID_LIST
+#endif /* defined(LOCAL_PEERPID) && defined(HAVE_LIBPROC_H) */
+
 #include "tracelib.h"
 
+#include "Pextlib.h"
+
 #include "strlcat.h"
 
 #ifdef HAVE_TRACEMODE_SUPPORT
@@ -83,6 +90,14 @@ size_t strlcpy(char *dst, const char *src, size_t size) {
 }
 #endif
 
+#ifdef HAVE_PEERPID_LIST
+static bool peerpid_list_enqueue(int sock, pid_t pid);
+static pid_t peerpid_list_dequeue(int sock);
+static pid_t peerpid_list_get(int sock, const char **progname);
+static void peerpid_list_walk(bool (*callback)(int sock, pid_t pid, const char *progname));
+#endif /* defined(HAVE_PEERPID_LIST) */
+
+
 static char *name;
 static char *sandbox;
 static size_t sandboxLength;
@@ -94,8 +109,28 @@ static int kq = -1;
 static int selfpipe[2];
 static int enable_fence = 0;
 static Tcl_Interp *interp;
-static pthread_mutex_t sock_mutex = PTHREAD_MUTEX_INITIALIZER;
-static int cleanuping = 0;
+
+/**
+ * Mutex that shall be acquired to exclusively lock checking and acting upon
+ * the value of kq, indicating whether the event loop has started. If it has
+ * started, shutdown of the event loop shall occur by writing to the write end
+ * of the selfpipe (which is non-blocking), which will in turn trigger the
+ * event loop termination and a signal on the evloop_signal condition variable
+ * when the loop has been terminated and it is safe to free the resources that
+ * were used by the loop.
+ *
+ * If kq is -1, the event loop has not been started and resources can
+ * immediately be free(3)d (under the lock to avoid concurrent set up of the
+ * event loop in a different thread).
+ */
+static pthread_mutex_t evloop_mutex = PTHREAD_MUTEX_INITIALIZER;
+
+/**
+ * Condition variable that shall be used to signal the end of the event loop
+ * after a termination signal has been sent to it via the write end of
+ * selfpipe. The associated mutex is evloop_mtx.
+ */
+static pthread_cond_t evloop_signal = PTHREAD_COND_INITIALIZER;
 
 static void send_file_map(int sock);
 static void dep_check(int sock, char *path);
@@ -106,13 +141,131 @@ typedef enum {
 } sandbox_violation_t;
 static void sandbox_violation(int sock, const char *path, sandbox_violation_t type);
 
-static void ui_warn(const char *format, ...) __printflike(1, 2);
-#if 0
-static void ui_info(const char *format, ...) __printflike(1, 2);
-#endif
-static void ui_error(const char *format, ...) __printflike(1, 2);
+#ifdef HAVE_PEERPID_LIST
+typedef struct _peerpid {
+    struct _peerpid *ppid_next;
+    char            *ppid_prog;
+    int              ppid_sock;
+    pid_t            ppid_pid;
+} peerpid_entry_t;
 
-#define MAX_SOCKETS (1024)
+static peerpid_entry_t *peer_list = NULL;
+
+/**
+ * Add a new entry to the list of PIDs of peers. Call this once for each
+ * accepted socket with the socket and the peer's PID.
+ *
+ * @param sock The new socket that was opened by the process with the given PID
+ *             and should be added to the list of peers.
+ * @param pid The PID of the new peer.
+ * @return boolean indicating success.
+ */
+static bool peerpid_list_enqueue(int sock, pid_t pid) {
+    char pathbuf[PROC_PIDPATHINFO_MAXSIZE];
+    const char *progname = "<unknown>";
+
+    peerpid_entry_t *ppid = malloc(sizeof(peerpid_entry_t));
+    if (!ppid) {
+        return false;
+    }
+
+    if (proc_pidpath(pid, pathbuf, sizeof(pathbuf))) {
+        progname = pathbuf;
+    }
+
+    ppid->ppid_prog = strdup(progname);
+    if (!ppid->ppid_prog) {
+        free(ppid);
+        return false;
+    }
+    ppid->ppid_sock = sock;
+    ppid->ppid_pid = pid;
+    ppid->ppid_next = peer_list;
+    peer_list = ppid;
+    return true;
+}
+
+/**
+ * Given a socket, dequeue a peer from the current list of peers. Use this when
+ * a socket is closed.
+ *
+ * @param sock The socket that is being closed and should be dequeued.
+ * @return The PID of the socket that has been dequeued, or (pid_t) -1
+ */
+static pid_t peerpid_list_dequeue(int sock) {
+    peerpid_entry_t **ref = &peer_list;
+    while (*ref) {
+        peerpid_entry_t *curr = *ref;
+        if (curr->ppid_sock == sock) {
+            // dequeue the element
+            *ref = curr->ppid_next;
+            pid_t pid = curr->ppid_pid;
+            free(curr->ppid_prog);
+            free(curr);
+            return pid;
+        }
+
+        ref = &curr->ppid_next;
+    }
+
+    return (pid_t) -1;
+}
+
+/**
+ * Return the peer PID given a socket.
+ *
+ * @param sock The socket for which the peer PID is needed.
+ * @param progname A pointer that will point to the string that holds the
+ *                 command line corresponding to the PID at the time of
+ *                 enqueuing. Set to NULL if not needed.
+ * @return The peer's PID or (pid_t) -1, if the socket could not be found in the list.
+ */
+static pid_t peerpid_list_get(int sock, const char **progname) {
+    peerpid_entry_t *curr = peer_list;
+    while (curr) {
+        if (curr->ppid_sock == sock) {
+            if (progname) {
+                *progname = curr->ppid_prog;
+            }
+            return curr->ppid_pid;
+        }
+
+        curr = curr->ppid_next;
+    }
+
+    return (pid_t) -1;
+}
+
+/**
+ * Walk the current list of (socket, peer PID) pairs and call a callback
+ * function for each pair.
+ *
+ * @param func Callback function to call for each tuple of socket, peer PID and
+ *             peer command line. The function should take an integer (the
+ *             socket), a pid_t (the peer's PID) and a const char * (the peer's
+ *             command line) and return a boolean (true, if the element should
+ *             be removed from the list, false otherwise). The callback must
+ *             not modify the list using peerpid_list_enqueue() or
+ *             peerpid_list_dequeue().
+ */
+static void peerpid_list_walk(bool (*callback)(int sock, pid_t pid, const char *progname)) {
+    peerpid_entry_t **ref = &peer_list;
+    while (*ref) {
+        peerpid_entry_t *curr = *ref;
+        if (callback(curr->ppid_sock, curr->ppid_pid, curr->ppid_prog)) {
+            // dequeue the element
+            *ref = curr->ppid_next;
+            free(curr->ppid_prog);
+            free(curr);
+            continue;
+        }
+
+        ref = &curr->ppid_next;
+    }
+}
+#endif /* defined(HAVE_PEERPID_LIST) */
+
+#define MAX_SOCKETS (64)
 #define BUFSIZE     (4096)
 
 /**
@@ -141,6 +294,38 @@ static void answer(int sock, const char *buf) {
     answer_s(sock, buf, (uint32_t) strlen(buf));
 }
 
+/**
+ * Closes the two sockets given in \a p and sets their values to -1.
+ */
+static void pipe_cleanup(int p[2]) {
+    for (size_t i = 0; i < 2; ++i) {
+        if (p[i] != -1) {
+            close(p[i]);
+            p[i] = -1;
+        }
+    }
+}
+
+/**
+ * Helper function to simplify error handling. Converts the error indicated by
+ * \a msg, appended with a string representation of the UNIX error \a errno
+ * into a Tcl error by setting up the result of the Tcl interpreter \a interp
+ * accordingly.
+ *
+ * Returns TCL_ERROR to be used as the return value of the caller.
+ */
+static int error2tcl(const char *msg, int err, Tcl_Interp *interp) {
+    Tcl_SetErrno(err);
+    Tcl_ResetResult(interp);
+    if (err != 0) {
+        Tcl_AppendResult(interp, msg, (char *) Tcl_PosixError(interp), NULL);
+    } else {
+        Tcl_AppendResult(interp, msg, NULL);
+    }
+
+    return TCL_ERROR;
+}
+
 /**
  * Sets the path of the tracelib unix socket where darwintrace should attempt
  * to connect to. This path should be specific to the port being installed.
@@ -298,7 +483,7 @@ static int TracelibSetSandboxCmd(Tcl_Interp *interp, int objc, Tcl_Obj *CONST ob
 static int process_line(int sock) {
     char *f;
     char buf[BUFSIZE];
-    uint32_t len;
+    uint32_t len = 0;
     ssize_t ret;
 
     if ((ret = recv(sock, &len, sizeof(len), MSG_WAITALL)) != sizeof(len)) {
@@ -307,13 +492,17 @@ static int process_line(int sock) {
         } else if (ret == 0) {
             /* this usually means the socket was closed by the remote side */
         } else {
-            fprintf(stderr, "tracelib: partial data received: expected %ld, but got %ld on socket %d\n", (unsigned long) sizeof(len), (unsigned long) ret, sock);
+            fprintf(stderr, "tracelib: partial data received: expected %zu, but got %zd on socket %d\n", sizeof(len), ret, sock);
         }
         return 0;
     }
 
     if (len > BUFSIZE - 1) {
-        fprintf(stderr, "tracelib: transfer too large: %ld bytes sent, but buffer holds %d on socket %d\n", (unsigned long) len, (int) (BUFSIZE - 1), sock);
+        pid_t pid = (pid_t) -1;
+#ifdef HAVE_PEERPID_LIST
+        pid = peerpid_list_get(sock, NULL);
+#endif
+        fprintf(stderr, "tracelib: transfer too large: %" PRIu32 " bytes sent, but buffer holds %d on socket %d from pid %ld\n", len, BUFSIZE - 1, sock, (unsigned long) pid);
         return 0;
     }
 
@@ -321,7 +510,7 @@ static int process_line(int sock) {
         if (ret < 0) {
             perror("tracelib: recv");
         } else {
-            fprintf(stderr, "tracelib: partial data received: expected %ld, but got %ld on socket %d\n", (unsigned long) len, (unsigned long) ret, sock);
+            fprintf(stderr, "tracelib: partial data received: expected %" PRIu32 ", but got %zd on socket %d\n", len, ret, sock);
         }
         return 0;
     }
@@ -377,14 +566,14 @@ static void send_file_map(int sock) {
  * \param[in] path the offending path to be passed to the callback
  */
 static void sandbox_violation(int sock UNUSED, const char *path, sandbox_violation_t type) {
-    Tcl_SetVar(interp, "path", path, 0);
+    Tcl_SetVar(interp, "_sandbox_viol_path", path, 0);
     int retVal = TCL_OK;
     switch (type) {
         case SANDBOX_VIOLATION:
-            retVal = Tcl_Eval(interp, "slave_add_sandbox_violation $path");
+            retVal = Tcl_Eval(interp, "slave_add_sandbox_violation ${_sandbox_viol_path}");
             break;
         case SANDBOX_UNKNOWN:
-            retVal = Tcl_Eval(interp, "slave_add_sandbox_unknown $path");
+            retVal = Tcl_Eval(interp, "slave_add_sandbox_unknown ${_sandbox_viol_path}");
             break;
     }
 
@@ -392,7 +581,7 @@ static void sandbox_violation(int sock UNUSED, const char *path, sandbox_violati
         fprintf(stderr, "Error evaluating Tcl statement to add sandbox violation: %s\n", Tcl_GetStringResult(interp));
     }
 
-    Tcl_UnsetVar(interp, "path", 0);
+    Tcl_UnsetVar(interp, "_sandbox_viol_path", 0);
 }
 
 /**
@@ -420,7 +609,7 @@ static void dep_check(int sock, char *path) {
     reg_error error;
 
     if (NULL == (reg = registry_for(interp, reg_attached))) {
-        ui_error("%s", Tcl_GetStringResult(interp));
+        ui_error(interp, "%s", Tcl_GetStringResult(interp));
         /* send unexpected output to make the build fail */
         answer(sock, "#");
     }
@@ -438,7 +627,7 @@ static void dep_check(int sock, char *path) {
     /* find the port's name to compare with out list */
     if (!reg_entry_propget(&entry, "name", &port, &error)) {
         /* send unexpected output to make the build fail */
-        ui_error("%s", error.description);
+        ui_error(interp, "%s", error.description);
         answer(sock, "#");
     }
 
@@ -455,364 +644,328 @@ static void dep_check(int sock, char *path) {
     answer(sock, "!");
 }
 
-__printflike(2, 0)
-static void ui_msg(const char *severity, const char *format, va_list va) {
-    char buf[1024], tclcmd[32];
-
-    vsnprintf(buf, sizeof(buf), format, va);
-
-    snprintf(tclcmd, sizeof(tclcmd), "ui_%s $warn", severity);
-
-    Tcl_SetVar(interp, "warn", buf, 0);
-    if (TCL_OK != Tcl_Eval(interp, tclcmd)) {
-        fprintf(stderr, "Error evaluating tcl statement `%s': %s\n", tclcmd, Tcl_GetStringResult(interp));
-    }
-    Tcl_UnsetVar(interp, "warn", 0);
-
-}
-
-__printflike(1, 2)
-static void ui_warn(const char *format, ...) {
-    va_list va;
-
-    va_start(va, format);
-    ui_msg("warn", format, va);
-    va_end(va);
-}
-
-#if 0
-__printflike(1, 2)
-static void ui_info(const char *format, ...) {
-    va_list va;
-
-    va_start(va, format);
-    ui_msg("info", format, va);
-    va_end(va);
-}
-#endif
-
-__printflike(1, 2)
-static void ui_error(const char *format, ...) {
-    va_list va;
-    va_start(va, format);
-    ui_msg("error", format, va);
-    va_end(va);
-}
-
 static int TracelibOpenSocketCmd(Tcl_Interp *in) {
     struct sockaddr_un sun;
-    struct rlimit rl;
 
-    cleanuping = 0;
-
-    pthread_mutex_lock(&sock_mutex);
     if (-1 == (sock = socket(PF_LOCAL, SOCK_STREAM, 0))) {
-        Tcl_SetErrno(errno);
-        Tcl_ResetResult(interp);
-        Tcl_AppendResult(interp, "socket: ", (char *) Tcl_PosixError(interp), NULL);
-        pthread_mutex_unlock(&sock_mutex);
-        return TCL_ERROR;
-    }
-    pthread_mutex_unlock(&sock_mutex);
-
-    interp = in;
-
-    /* raise the limit of open files to the maximum from the default soft limit
-     * of 256 */
-    if (getrlimit(RLIMIT_NOFILE, &rl) == -1) {
-        ui_warn("getrlimit failed (%d), skipping setrlimit", errno);
-    } else {
-#ifdef OPEN_MAX
-        if (rl.rlim_max > OPEN_MAX) {
-            rl.rlim_max = OPEN_MAX;
-        }
-#endif
-        rl.rlim_cur = rl.rlim_max;
-        if (setrlimit(RLIMIT_NOFILE, &rl) == -1) {
-            ui_warn("setrlimit failed (%d)", errno);
-        }
+        return error2tcl("socket: ", errno, in);
     }
 
     sun.sun_family = AF_UNIX;
     strlcpy(sun.sun_path, name, sizeof(sun.sun_path));
 
     if (-1 == (bind(sock, (struct sockaddr *) &sun, sizeof(sun)))) {
-        Tcl_SetErrno(errno);
-        Tcl_ResetResult(interp);
-        Tcl_AppendResult(interp, "bind: ", (char *) Tcl_PosixError(interp), NULL);
+        int err = errno;
         close(sock);
         sock = -1;
-        return TCL_ERROR;
+        return error2tcl("bind: ", err, in);
     }
 
-    if (-1 == listen(sock, 32)) {
-        Tcl_SetErrno(errno);
-        Tcl_ResetResult(interp);
-        Tcl_AppendResult(interp, "listen: ", (char *) Tcl_PosixError(interp), NULL);
+    if (-1 == listen(sock, SOMAXCONN)) {
+        int err = errno;
         close(sock);
         sock = -1;
-        return TCL_ERROR;
+        return error2tcl("bind: ", err, in);
     }
 
+    // keep a reference to the interpreter that opened the socket
+    interp = in;
+
     return TCL_OK;
 }
 
+#ifdef HAVE_PEERPID_LIST
+/**
+ * Callback to be passed to peerpid_list_walk(). Closes the open sockets and
+ * sends SIGTERM to the associated processes. Leaves the list unmodified.
+ */
+static bool close_and_send_sigterm(int sock UNUSED, pid_t pid, const char *progname) {
+    ui_warn(interp, "Sending SIGTERM to process %ld: %s", (unsigned long) pid, progname);
+    kill(pid, SIGTERM);
+
+    // keep the elements in the list
+    return false;
+}
+
+/**
+ * Callback to be passed to peerpid_list_walk(). Sends SIGKILL to the processes
+ * and deletes the elements from the list.
+ */
+static bool send_sigkill_and_free(int sock, pid_t pid, const char *progname UNUSED) {
+    close(sock);
+    kill(pid, SIGKILL);
+
+    // remove the elements from the list
+    return true;
+}
+#endif
+
 /* create this on heap rather than stack, due to its rather large size */
 static struct kevent res_kevents[MAX_SOCKETS];
 
 static int TracelibRunCmd(Tcl_Interp *in) {
     struct kevent kev;
+    int retval = TCL_ERROR;
     int flags;
-    int oldsock;
     int opensockcount = 0;
+    bool break_eventloop = false;
+
+    pthread_mutex_lock(&evloop_mutex);
+    /* bring all variables into a defined state so the cleanup code can be
+     * called from anywhere */
+    selfpipe[0] = -1;
+    selfpipe[1] = -1;
+    kq = -1;
 
-    pthread_mutex_lock(&sock_mutex);
     if (-1 == (kq = kqueue())) {
-        Tcl_SetErrno(errno);
-        Tcl_ResetResult(in);
-        Tcl_AppendResult(in, "kqueue: ", (char *) Tcl_PosixError(in), NULL);
-        return TCL_ERROR;
+        error2tcl("kqueue: ", errno, in);
+        goto error_locked;
     }
 
     if (sock != -1) {
-        oldsock = sock;
-
         /* mark listen socket non-blocking in order to prevent a race condition
          * that would occur between kevent(2) and accept(2), if a incoming
          * connection is aborted before it is accepted. Using a non-blocking
          * accept(2) prevents the problem.*/
-        flags = fcntl(oldsock, F_GETFL, 0);
-        if (-1 == fcntl(oldsock, F_SETFL, flags | O_NONBLOCK)) {
-            Tcl_SetErrno(errno);
-            Tcl_ResetResult(in);
-            Tcl_AppendResult(in, "fcntl(F_SETFL, += O_NONBLOCK): ", (char *) Tcl_PosixError(in), NULL);
-            pthread_mutex_unlock(&sock_mutex);
-            return TCL_ERROR;
+        flags = fcntl(sock, F_GETFL, 0);
+        if (-1 == fcntl(sock, F_SETFL, flags | O_NONBLOCK)) {
+            error2tcl("fcntl(F_SETFL, += O_NONBLOCK): ", errno, in);
+            goto error_locked;
         }
 
         /* register the listen socket in the kqueue */
-        EV_SET(&kev, oldsock, EVFILT_READ, EV_ADD | EV_RECEIPT, 0, 0, NULL);
+        EV_SET(&kev, sock, EVFILT_READ, EV_ADD | EV_RECEIPT, 0, 0, NULL);
         if (1 != kevent(kq, &kev, 1, &kev, 1, NULL)) {
-            Tcl_SetErrno(errno);
-            Tcl_ResetResult(in);
-            Tcl_AppendResult(in, "kevent (listen socket): ", (char *) Tcl_PosixError(in), NULL);
-            close(kq);
-            pthread_mutex_unlock(&sock_mutex);
-            return TCL_ERROR;
+            error2tcl("kevent (listen socket): ", errno, in);
+            goto error_locked;
         }
         /* kevent(2) on EV_RECEIPT: When passed as input, it forces EV_ERROR to
          * always be returned. When a filter is successfully added, the data field
          * will be zero. */
         if ((kev.flags & EV_ERROR) == 0 || ((kev.flags & EV_ERROR) > 0 && kev.data != 0)) {
-            Tcl_SetErrno(kev.data);
-            Tcl_ResetResult(in);
-            Tcl_AppendResult(in, "kevent (listen socket receipt): ", (char *) Tcl_PosixError(in), NULL);
-            close(kq);
-            pthread_mutex_unlock(&sock_mutex);
-            return TCL_ERROR;
+            error2tcl("kevent (listen socket receipt): ", kev.data, in);
+            goto error_locked;
         }
 
 
         /* use the self-pipe trick to trigger returning from kevent(2) when
          * tracelib closesocket is called. */
         if (-1 == pipe(selfpipe)) {
-            Tcl_SetErrno(errno);
-            Tcl_ResetResult(in);
-            Tcl_AppendResult(in, "pipe: ", (char *) Tcl_PosixError(in), NULL);
-            pthread_mutex_unlock(&sock_mutex);
-            return TCL_ERROR;
+            error2tcl("pipe: ", errno, in);
+            goto error_locked;
         }
 
         /* mark the write side of the pipe non-blocking */
         flags = fcntl(selfpipe[1], F_GETFL, 0);
         if (-1 == fcntl(selfpipe[1], F_SETFL, flags | O_NONBLOCK)) {
-            Tcl_SetErrno(errno);
-            Tcl_ResetResult(in);
-            Tcl_AppendResult(in, "fcntl(F_SETFL, += O_NONBLOCK): ", (char *) Tcl_PosixError(in), NULL);
-            pthread_mutex_unlock(&sock_mutex);
-            return TCL_ERROR;
+            error2tcl("fcntl(F_SETFL, += O_NONBLOCK): ", errno, in);
+            goto error_locked;
         }
 
         /* wait for the user event on the listen socket, as sent by CloseCmd as
          * deathpill */
         EV_SET(&kev, selfpipe[0], EVFILT_READ, EV_ADD | EV_RECEIPT, 0, 0, NULL);
         if (1 != kevent(kq, &kev, 1, &kev, 1, NULL)) {
-            Tcl_SetErrno(errno);
-            Tcl_ResetResult(in);
-            Tcl_AppendResult(in, "kevent (selfpipe): ", (char *) Tcl_PosixError(in), NULL);
-            close(kq);
-            pthread_mutex_unlock(&sock_mutex);
-            return TCL_ERROR;
+            error2tcl("kevent (selfpipe): ", errno, in);
+            goto error_locked;
         }
         /* kevent(2) on EV_RECEIPT: When passed as input, it forces EV_ERROR to
          * always be returned. When a filter is successfully added, the data field
          * will be zero. */
         if ((kev.flags & EV_ERROR) == 0 || ((kev.flags & EV_ERROR) > 0 && kev.data != 0)) {
-            Tcl_SetErrno(kev.data);
-            Tcl_ResetResult(in);
-            Tcl_AppendResult(in, "kevent (selfpipe receipt): ", (char *) Tcl_PosixError(in), NULL);
-            close(kq);
-            pthread_mutex_unlock(&sock_mutex);
-            return TCL_ERROR;
+            error2tcl("kevent (selfpipe receipt): ", kev.data, in);
+            goto error_locked;
         }
     }
-    pthread_mutex_unlock(&sock_mutex);
+    pthread_mutex_unlock(&evloop_mutex);
 
-    while (sock != -1 && !cleanuping) {
+    while (sock != -1 && !break_eventloop) {
         int keventstatus;
-        int i;
+        bool incoming = false;
 
         /* run kevent(2) until new activity is available */
         do {
             if (-1 == (keventstatus = kevent(kq, NULL, 0, res_kevents, MAX_SOCKETS, NULL))) {
-                Tcl_SetErrno(errno);
-                Tcl_ResetResult(in);
-                Tcl_AppendResult(in, "kevent (main loop): ", (char *) Tcl_PosixError(in), NULL);
-                close(kq);
-                return TCL_ERROR;
+                error2tcl("kevent (main loop): ", errno, in);
+                goto error_unlocked;
             }
         } while (keventstatus == 0);
 
-        for (i = 0; i < keventstatus; ++i) {
+        for (int i = 0; i < keventstatus; ++i) {
             /* handle traffic on the selfpipe */
             if ((int) res_kevents[i].ident == selfpipe[0]) {
-                pthread_mutex_lock(&sock_mutex);
-                close(selfpipe[0]);
-                close(selfpipe[1]);
-                selfpipe[0] = -1;
-                selfpipe[1] = -1;
-                pthread_mutex_unlock(&sock_mutex);
-                break;
-            }
-
-            /* the control socket has activity – we might have a new
-             * connection. We use a copy of sock here, because sock might have
-             * been set to -1 by the close command */
-            if ((int) res_kevents[i].ident == oldsock) {
-                int s;
-
-                /* handle error conditions */
-                if ((res_kevents[i].flags & (EV_ERROR | EV_EOF)) > 0) {
-                    if (cleanuping) {
-                        break;
-                    }
-                    Tcl_ResetResult(in);
-                    Tcl_SetResult(in, "control socket closed", NULL);
-                    close(kq);
-                    return TCL_ERROR;
-                }
-
-                /* else: new connection attempt(s) */
-                for (;;) {
-                    if (-1 == (s = accept(sock, NULL, NULL))) {
-                        if (cleanuping) {
-                            break;
-                        }
-                        if (errno == EWOULDBLOCK) {
-                            break;
-                        }
-                        Tcl_SetErrno(errno);
-                        Tcl_ResetResult(in);
-                        Tcl_AppendResult(in, "accept: ", (char *) Tcl_PosixError(in), NULL);
-                        close(kq);
-                        return TCL_ERROR;
-                    }
-
-                    flags = fcntl(s, F_GETFL, 0);
-                    if (-1 == fcntl(s, F_SETFL, flags & ~O_NONBLOCK)) {
-                        ui_warn("tracelib: couldn't mark socket as blocking");
-                        close(s);
-                        continue;
-                    }
-
-                    /* register the new socket in the kqueue */
-                    EV_SET(&kev, s, EVFILT_READ, EV_ADD | EV_RECEIPT, 0, 0, NULL);
-                    if (1 != kevent(kq, &kev, 1, &kev, 1, NULL)) {
-                        ui_warn("tracelib: error adding socket to kqueue");
-                        close(s);
-                        continue;
-                    }
-                    /* kevent(2) on EV_RECEIPT: When passed as input, it forces EV_ERROR to
-                     * always be returned. When a filter is successfully added, the data field
-                     * will be zero. */
-                    if ((kev.flags & EV_ERROR) == 0 || ((kev.flags & EV_ERROR) > 0 && kev.data != 0)) {
-                        ui_warn("tracelib: error adding socket to kqueue");
-                        close(s);
-                        continue;
-                    }
-
-                    opensockcount++;
-                }
-
-                if (cleanuping) {
-                    break;
-                }
-            } else {
+                /* traffic on the selfpipe means we should clean up */
+                break_eventloop = true;
+                /* finish processing this batch */
+                continue;
+            } else if ((int) res_kevents[i].ident != sock) {
                 /* if the socket is to be closed, or */
                 if ((res_kevents[i].flags & (EV_EOF | EV_ERROR)) > 0
                     /* new data is available, and its processing tells us to
                      * close the socket */
                     || (!process_line(res_kevents[i].ident))) {
-                    /* an error occured or process_line suggested closing this
-                     * socket */
-                    close(res_kevents[i].ident);
-                    /* closing the socket will automatically remove it from the
-                     * kqueue :) */
-                    opensockcount--;
+                        /* an error occured or process_line suggested closing
+                         * this socket */
+                        close(res_kevents[i].ident);
+                        /* closing the socket will automatically remove it from the
+                         * kqueue :) */
+                        opensockcount--;
+
+#ifdef HAVE_PEERPID_LIST
+                        if (peerpid_list_dequeue(res_kevents[i].ident) == (pid_t) -1) {
+                            fprintf(stderr, "tracelib: didn't find PID for closed socket %d\n", (int) res_kevents[i].ident);
+                        }
+#endif
                 }
+            } else {
+                /* the control socket has activity – we might have a new
+                 * connection. */
+
+                /* handle error conditions */
+                if ((res_kevents[i].flags & (EV_ERROR | EV_EOF)) > 0) {
+                    error2tcl("control socket closed", 0, in);
+                    goto error_unlocked;
+                }
+
+                /* delay processing, process data on existing sockets first */
+                incoming = true;
+            }
+        }
+
+        if (incoming) {
+            /* new connection attempt(s) */
+            for (;;) {
+                int s;
+
+                if (-1 == (s = accept(sock, NULL, NULL))) {
+                    if (errno == EWOULDBLOCK) {
+                        break;
+                    }
+
+                    error2tcl("accept: ", errno, in);
+                    goto error_unlocked;
+                }
+
+                flags = fcntl(s, F_GETFL, 0);
+                if (-1 == fcntl(s, F_SETFL, flags & ~O_NONBLOCK)) {
+                    ui_warn(interp, "tracelib: couldn't mark socket as blocking");
+                    close(s);
+                    continue;
+                }
+
+                /* register the new socket in the kqueue */
+                EV_SET(&kev, s, EVFILT_READ, EV_ADD | EV_RECEIPT, 0, 0, NULL);
+                if (1 != kevent(kq, &kev, 1, &kev, 1, NULL)) {
+                    ui_warn(interp, "tracelib: error adding socket to kqueue");
+                    close(s);
+                    continue;
+                }
+                /* kevent(2) on EV_RECEIPT: When passed as input, it forces EV_ERROR to
+                 * always be returned. When a filter is successfully added, the data field
+                 * will be zero. */
+                if ((kev.flags & EV_ERROR) == 0 || ((kev.flags & EV_ERROR) > 0 && kev.data != 0)) {
+                    ui_warn(interp, "tracelib: error adding socket to kqueue (receipt)");
+                    close(s);
+                    continue;
+                }
+
+#ifdef HAVE_PEERPID_LIST
+                pid_t peer_pid = (pid_t) -1;
+                socklen_t peer_pid_len = sizeof(peer_pid);
+                if (getsockopt(s, SOL_LOCAL, LOCAL_PEERPID, &peer_pid, &peer_pid_len) == 0) {
+                    // We found a PID for the remote side
+                    peerpid_list_enqueue(s, peer_pid);
+                } else {
+                    // Error occured, process has probably already terminated
+                    close(s);
+                    continue;
+                }
+#endif
+                opensockcount++;
             }
         }
     }
 
-    /* NOTE: We aren't necessarily closing all client sockets here! */
-    if (opensockcount > 0) {
-        fprintf(stderr, "tracelib: %d open sockets will leak at end of runcmd\n", opensockcount);
-    }
-    pthread_mutex_lock(&sock_mutex);
-    close(kq);
-    kq = -1;
-    pthread_mutex_unlock(&sock_mutex);
+    retval = TCL_OK;
 
-    return TCL_OK;
+error_unlocked:
+    pthread_mutex_lock(&evloop_mutex);
+error_locked:
+    // Close remainig sockets to avoid dangling processes
+    if (opensockcount > 0) {
+#ifdef HAVE_PEERPID_LIST
+        ui_warn(interp, "tracelib: %d open sockets leaking at end of runcmd, closing, sending SIGTERM and SIGKILL", opensockcount);
+        peerpid_list_walk(close_and_send_sigterm);
+        peerpid_list_walk(send_sigkill_and_free);
+#else
+        ui_warn(interp, "tracelib: %d open sockets leaking at end of runcmd", opensockcount);
+#endif
+    }
+
+    // cleanup selfpipe and set it to -1
+    pipe_cleanup(selfpipe);
+
+    // close kqueue(2) socket
+    if (kq != -1) {
+        close(kq);
+        kq = -1;
+    }
+
+    pthread_mutex_unlock(&evloop_mutex);
+    // wake up any waiting threads in TracelibCloseSocketCmd
+    pthread_cond_broadcast(&evloop_signal);
+
+    return retval;
 }
 
 static int TracelibCleanCmd(Tcl_Interp *interp UNUSED) {
-#define safe_free(x) do{free(x); x=0;}while(0);
-    cleanuping = 1;
-    pthread_mutex_lock(&sock_mutex);
+#define safe_free(x) do{ \
+        free(x); \
+        x = NULL; \
+    } while(0);
+
     if (sock != -1) {
         close(sock);
         sock = -1;
     }
-    pthread_mutex_unlock(&sock_mutex);
+
     if (name) {
         unlink(name);
         safe_free(name);
     }
-    if (depends) {
-        safe_free(depends);
-    }
+
+    safe_free(depends);
+
     enable_fence = 0;
-#undef safe_free
     return TCL_OK;
+
+#undef safe_free
 }
 
 static int TracelibCloseSocketCmd(Tcl_Interp *interp UNUSED) {
-    cleanuping = 1;
-    pthread_mutex_lock(&sock_mutex);
-    if (sock != -1) {
-        close(sock);
-        sock = -1;
+    pthread_mutex_lock(&evloop_mutex);
+    if (kq != -1 && selfpipe[1] != -1) {
+        /* We know the pipes have been created because kq != -1 and we have the
+         * lock. We don't have to check for errors, because none should occur
+         * but when the pipe is full, which we wouldn't care about. */
+        write(selfpipe[1], "!", 1);
 
-        if (kq != -1) {
-            /* We know the pipes have been created because kq != -1 and we have
-             * the lock. We don't have to check for errors, because none should
-             * occur but when the pipe is full, which we wouldn't care about.
-             * */
-            write(selfpipe[1], "!", 1);
+        /* Wait for the kqueue event loop to terminate. We must not return
+         * earlier than that because the next call will be to tracelib clean,
+         * and that frees up memory that would be used by the event loop
+         * otherwise. */
+        pthread_cond_wait(&evloop_signal, &evloop_mutex);
+    } else {
+        /* The kqueue(2) loop isn't running yet, so we can just close the
+         * socket and make sure it stays closed. In this situation, the kqueue
+         * will not be created. */
+        if (sock != -1) {
+            close(sock);
+            sock = -1;
         }
     }
-    pthread_mutex_unlock(&sock_mutex);
+    pthread_mutex_unlock(&evloop_mutex);
+
     return TCL_OK;
 }
 
