@@ -1,6 +1,5 @@
 # et:ts=4
 # port.tcl
-# $Id: port.tcl 79597 2011-06-19 20:59:11Z jmr@macports.org $
 #
 # Copyright (c) 2002 Apple Inc.
 # All rights reserved.
@@ -32,6 +31,37 @@
 # standard package load
 package provide port 1.0
 
+# catch wrapper shared with macports1.0
+# aliasing it in doesn't work right because of uplevel use
+package require mpcommon 1.0
+
+# Provide a callback registration mechanism for port subpackages. This needs to
+# be done _before_ loading the subpackages.
+namespace eval port {
+	variable _callback_list [list]
+
+	# Append a new procedure to a list of callbacks to be called when
+	# port::run_callbacks is called from macports1.0 after evaluating
+	# a Portfile
+	proc register_callback {callback} {
+		variable _callback_list
+		lappend _callback_list ${callback}
+	}
+	
+	# Run the callbacks registered in the callback list. Called from
+	# macports1.0 in the child interpreter after evaluating the Portfile and
+	# the variants. Clears the list of callbacks.
+	proc run_callbacks {} {
+		variable _callback_list
+		foreach callback ${_callback_list} {
+			ui_debug "Running callback ${callback}"
+			${callback}
+			ui_debug "Finished running callback ${callback}"
+		}
+		set _callback_list [list]
+	}
+}
+
 package require mp_package 1.0
 package require portmain 1.0
 package require portdepends 1.0
@@ -49,12 +79,13 @@ package require portdeactivate 1.0
 package require portclean 1.0
 package require porttest 1.0
 package require portlint 1.0
-package require portsubmit 1.0
 package require porttrace 1.0
 package require portdistcheck 1.0
 package require portlivecheck 1.0
 package require portmirror 1.0
 package require portload 1.0
 package require portunload 1.0
+package require portreload 1.0
 
 package require portdistfiles 1.0
+package require portsandbox 1.0
