@@ -1,6 +1,32 @@
-# et:ts=4
-# portlint.tcl
-# $Id$
+# -*- coding: utf-8; mode: tcl; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- vim:fenc=utf-8:filetype=tcl:et:sw=4:ts=4:sts=4
+#
+# Copyright (c) 2007 - 2016 The MacPorts Project
+# All rights reserved.
+#
+# Redistribution and use in source and binary forms, with or without
+# modification, are permitted provided that the following conditions
+# are met:
+# 1. Redistributions of source code must retain the above copyright
+#    notice, this list of conditions and the following disclaimer.
+# 2. Redistributions in binary form must reproduce the above copyright
+#    notice, this list of conditions and the following disclaimer in the
+#    documentation and/or other materials provided with the distribution.
+# 3. Neither the name of The MacPorts Project nor the names of its contributors
+#    may be used to endorse or promote products derived from this software
+#    without specific prior written permission.
+# 
+# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
+# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+# POSSIBILITY OF SUCH DAMAGE.
+#
 
 package provide portlint 1.0
 package require portutil 1.0
@@ -53,6 +79,7 @@ set lint_optional [list \
     "distname" \
     "use_automake" \
     "use_autoconf" \
+    "use_autoreconf" \
     "use_configure" \
     ]
 
@@ -258,6 +285,19 @@ proc portlint::lint_main {args} {
             incr warnings
         }
 
+        if {[regexp {compiler\.blacklist(?:-[a-z]+)?\s.*(["{]\S+(?:\s+\S+){2,}["}])} $line -> blacklist] && ![info exists portgroups(compiler_blacklist_versions)]} {
+            ui_error "Line $lineno uses compiler.blacklist entry $blacklist which requires the compiler_blacklist_versions portgroup which has not been included"
+            incr errors
+        }
+
+        if {[regexp {(^.*)(\meval\s+)(.*)(\[glob\M)(.*$)} $line -> match_before match_eval match_between match_glob match_after]} {
+            ui_warn "Line $lineno should use the expansion operator instead of the eval procedure. Change"
+            ui_warn "$line"
+            ui_warn "to"
+            ui_warn "$match_before$match_between{*}$match_glob$match_after"
+            incr warnings
+        }
+
         # Check for hardcoded version numbers
         if {$nitpick} {
             # Support for skipping checksums lines
@@ -290,6 +330,7 @@ proc portlint::lint_main {args} {
                 && ![regexp {^\s*reinplace} $line]
                 && ![regexp {^\s*system.*\Wsed\W} $line]} {
             ui_error "Line $lineno hardcodes /opt/local, use \${prefix} instead"
+            incr errors
         }
 
         ### TODO: more checks to Portfile syntax
@@ -302,22 +343,22 @@ proc portlint::lint_main {args} {
     global os.platform os.arch os.version version revision epoch \
            description long_description platforms categories all_variants \
            maintainers license homepage master_sites checksums patchfiles \
-           depends_fetch depends_extract depends_lib depends_build \
-           depends_run distfiles fetch.type lint_portsystem lint_platforms \
-           lint_required lint_optional
+           depends_fetch depends_extract depends_lib depends_build depends_run \
+           depends_test distfiles fetch.type lint_portsystem lint_platforms \
+           lint_required lint_optional replaced_by conflicts
     set portarch [get_canonical_archs]
 
-    if (!$seen_portsystem) {
+    if {!$seen_portsystem} {
         ui_error "Didn't find PortSystem specification"
         incr errors
-    }  elseif {$portsystem != $lint_portsystem} {
+    }  elseif {$portsystem ne $lint_portsystem} {
         ui_error "Unknown PortSystem: $portsystem"
         incr errors
     } else {
         ui_info "OK: Found PortSystem $portsystem"
     }
 
-    if ($seen_portgroup) {
+    if {$seen_portgroup} {
         # Using a PortGroup is optional
         foreach {portgroup portgroupversion} [array get portgroups] {
             if {![file exists [getportresourcepath $porturl "port1.0/group/${portgroup}-${portgroupversion}.tcl"]]} {
@@ -332,7 +373,7 @@ proc portlint::lint_main {args} {
     foreach req_var $lint_required {
 
         if {$req_var eq "master_sites"} {
-            if {${fetch.type} != "standard"} {
+            if {${fetch.type} ne "standard"} {
                 ui_info "OK: $req_var not required for fetch.type ${fetch.type}"
                 continue
             }
@@ -366,7 +407,7 @@ proc portlint::lint_main {args} {
 
     if {[info exists platforms]} {
         foreach platform $platforms {
-            if {[lsearch -exact $lint_platforms $platform] == -1} {
+            if {$platform ni $lint_platforms} {
                 ui_error "Unknown platform: $platform"
                 incr errors
             } else {
@@ -413,7 +454,7 @@ proc portlint::lint_main {args} {
 
             if {![info exists variantdesc] || $variantdesc eq ""} {
                 # don't warn about missing descriptions for global variants
-                if {[lsearch -exact $local_variants $variantname] != -1 &&
+                if {$variantname in $local_variants &&
                     [variant_desc $porturl $variantname] eq ""} {
                     ui_warn "Variant $variantname does not have a description"
                     incr warnings
@@ -432,7 +473,7 @@ proc portlint::lint_main {args} {
             foreach vconflict [ditem_key $variant conflicts] {
                 set exists 0
                 foreach v $all_variants {
-                    if {$vconflict == [ditem_key $v name]} {
+                    if {$vconflict eq [ditem_key $v name]} {
                         set exists 1
                         break
                     }
@@ -455,16 +496,28 @@ proc portlint::lint_main {args} {
     }
 
     set all_depends {}
-    if {[info exists depends_fetch]} { eval "lappend all_depends $depends_fetch" }
-    if {[info exists depends_extract]} { eval "lappend all_depends $depends_extract" }
-    if {[info exists depends_lib]} { eval "lappend all_depends $depends_lib" }
-    if {[info exists depends_build]} { eval "lappend all_depends $depends_build" }
-    if {[info exists depends_run]} { eval "lappend all_depends $depends_run" }
+    if {[info exists depends_fetch]} {
+        lappend all_depends {*}$depends_fetch
+    }
+    if {[info exists depends_extract]} {
+        lappend all_depends {*}$depends_extract
+    }
+    if {[info exists depends_lib]} {
+        lappend all_depends {*}$depends_lib
+    }
+    if {[info exists depends_build]} {
+        lappend all_depends {*}$depends_build
+    }
+    if {[info exists depends_run]} {
+        lappend all_depends {*}$depends_run
+    }
+    if {[info exists depends_test]} {
+        lappend all_depends {*}$depends_test
+    }
     foreach depspec $all_depends {
         set dep [lindex [split $depspec :] end]
         if {[catch {set res [mport_lookup $dep]} error]} {
-            global errorInfo
-            ui_debug "$errorInfo"
+            ui_debug $::errorInfo
             continue
         }
         if {$res eq ""} {
@@ -476,7 +529,7 @@ proc portlint::lint_main {args} {
     }
 
     # Check for multiple dependencies
-    foreach deptype {depends_extract depends_lib depends_build depends_run} {
+    foreach deptype {depends_extract depends_lib depends_build depends_run depends_test} {
         if {[info exists $deptype]} {
             array set depwarned {}
             foreach depspec [set $deptype] {
@@ -491,6 +544,43 @@ proc portlint::lint_main {args} {
         }
     }
 
+    if {[info exists replaced_by]} {
+        if {[regexp {[^[:alnum:]_.-]} $replaced_by]} {
+            ui_error "replaced_by should be a single port name, invalid value: $replaced_by"
+            incr errors
+        } else {
+            if {[catch {set res [mport_lookup $replaced_by]} error]} {
+                ui_debug $::errorInfo
+            }
+            if {$res eq ""} {
+                ui_error "replaced_by references unknown port: $replaced_by"
+                incr errors
+            } else {
+                ui_info "OK: replaced_by $replaced_by"
+            }
+        }
+    }
+
+    if {[info exists conflicts]} {
+        foreach cport $conflicts {
+            if {[regexp {[^[:alnum:]_.-]} $cport]} {
+                ui_error "conflicts lists invalid value, should be port name: $cport"
+                incr errors
+                continue
+            }
+            if {[catch {set res [mport_lookup $cport]} error]} {
+                ui_debug $::errorInfo
+                continue
+            }
+            if {$res eq ""} {
+                ui_error "conflicts references unknown port: $cport"
+                incr errors
+            } else {
+                ui_info "OK: conflicts $cport"
+            }
+        }
+    }
+
     if {[regexp "^(.+)nomaintainer(@macports.org)?(.+)$" $maintainers] } {
         ui_error "Using nomaintainer together with other maintainer"
         incr errors
@@ -502,14 +592,14 @@ proc portlint::lint_main {args} {
     }
 
     foreach addr $maintainers {
-        if {$addr == "nomaintainer@macports.org" ||
-                $addr == "openmaintainer@macports.org"} {
+        if {$addr eq "nomaintainer@macports.org" ||
+                $addr eq "openmaintainer@macports.org"} {
             ui_warn "Using full email address for no/open maintainer"
             incr warnings
         } elseif {[regexp "^(.+)@macports.org$" $addr -> localpart]} {
             ui_warn "Maintainer email address for $localpart includes @macports.org"
             incr warnings
-        } elseif {$addr == "darwinports@opendarwin.org"} {
+        } elseif {$addr eq "darwinports@opendarwin.org"} {
             ui_warn "Using legacy email address for no/open maintainer"
             incr warnings
         } elseif {[regexp "^(.+)@(.+)$" $addr -> localpart domain]} {
@@ -532,6 +622,7 @@ proc portlint::lint_main {args} {
             # space instead of hyphen
             if {[string is double -strict $test]} {
                 ui_error "Invalid license '${prev} ${test}': missing hyphen between ${prev} ${test}"
+                incr errors
 
             # missing hyphen
             } elseif {![string equal -nocase "X11" $test]} {
@@ -546,6 +637,7 @@ proc portlint::lint_main {args} {
                         set license_end [string index $subtest end]
                         if {"+" eq $license_end || [string is integer -strict $license_end]} {
                             ui_error "invalid license '${test}': missing hyphen before version"
+                            incr errors
                         }
                     }
                 }
@@ -554,12 +646,15 @@ proc portlint::lint_main {args} {
             if {[string equal -nocase "BSD-2" $test]} {
                 # BSD-2 => BSD
                 ui_error "Invalid license '${test}': use BSD instead"
+                incr errors
             } elseif {[string equal -nocase "BSD-3" $test]} {
                 # BSD-3 => BSD
                 ui_error "Invalid license '${test}': use BSD instead"
+                incr errors
             } elseif {[string equal -nocase "BSD-4" $test]} {
                 # BSD-4 => BSD-old
                 ui_error "Invalid license '${test}': use BSD-old instead"
+                incr errors
             }
 
             set prev $test
@@ -568,13 +663,13 @@ proc portlint::lint_main {args} {
     }
 
     # these checks are only valid for ports stored in the regular tree directories
-    if {[info exists category] && $portcatdir != $category} {
+    if {[info exists category] && $portcatdir ne $category} {
         ui_error "Portfile parent directory $portcatdir does not match primary category $category"
         incr errors
     } else {
         ui_info "OK: Portfile parent directory matches primary category"
     }
-    if {$portdir != $name} {
+    if {$portdir ne $name} {
         ui_error "Portfile directory $portdir does not match port name $name"
         incr errors
     } else {
